{ ******************************************************************
  Fast Fourier Transform (modified from Pascal program by Don Cross)
  ******************************************************************
  The program generates a time signal consisting of a large 200 Hz 
  sine wave added to a small 2000 Hz cosine wave, which is graphed
  on the screen (Press a key after you are done viewing each graph)

  Next, it performs the FFT and graphs the resulting complex
  frequency samples.

  Results are stored in the output file testfft.txt
  ****************************************************************** }

program testfft;

uses
  utypes, ufft, uinterv, uplot;

function F(T : Float) : Float;
begin
  F := Sin(200 * 2 * Pi * T) + 0.2 * Cos(2000 * 2 * Pi * T);
end;

const
  NumSamples   = 512;                        { Buffer size = power of 2 }
  SamplingRate = 22050;                      { Sampling rate (Hz) }
  MaxIndex     = NumSamples - 1;             { Max. array index }
  MidIndex     = NumSamples div 2;
  DT           = 1 / SamplingRate;           { Time unit (s) }
  DF           = SamplingRate / NumSamples;  { Frequency unit (Hz) }

var
  InArray, 
  OutArray : array[0..MaxIndex] of Complex;
  T, Freq  : array[0..MaxIndex] of Float;
  OutFile  : Text;
  I        : Integer;

procedure ListData(var DataArray : array[Lb..Ub : Integer] of Complex;  
                       Comment   : String);
var
  I : Integer;
begin
  WriteLn(OutFile, '*** ', Comment, ' ***');
  WriteLn(OutFile);
  WriteLn(OutFile, 'index':20, 'real':20, 'imag':20);
  
  for I := 0 to MaxIndex do
    WriteLn(OutFile, I:20,
            Re(DataArray[I]):20:5, Im(DataArray[I]):20:5);

  WriteLn(OutFile);
  WriteLn(OutFile, '------------------------------------------------------------------------');
  WriteLn(OutFile);
end;

procedure PlotData(var T : array[LbT..UbT : Integer] of Float; 
                   var Z : array[LbZ..UbZ : Integer] of Complex;  
                   Title : String);
var
  X : array[LbZ..UbZ] of Float;  { Real part of Z }
  I : Integer;                   { Loop variable }
  
  Xmin, Xmax, Xstep : Float;     { Ox scale }
  Ymin, Ymax, Ystep : Float;     { Oy scale }
  
begin
  if not InitGraphics(9, 2, '../../chr') then
    begin
      Writeln('Unable to set graphic mode!');
      Exit;
    end;

  SetWindow(15, 85, 15, 85, True);

  for I := 0 to MaxIndex do
    X[I] := Re(Z[I]);

  AutoScale(T, LinScale, Xmin, Xmax, Xstep);
  AutoScale(X, LinScale, Ymin, Ymax, Ystep);

  SetOxScale(LinScale, Xmin, Xmax, Xstep);
  SetOyScale(LinScale, Ymin, Ymax, Ystep);

  SetOxTitle('Time (s)');
  SetOyTitle('Amplitude');
  SetGraphTitle(Title);

  { Set point type to 0 so that only lines will be plotted }
  SetPointParam(1, 0, 1, 1);

  PlotOxAxis;
  PlotOyAxis;

  PlotGrid(BothGrid);

  WriteGraphTitle;

  PlotCurve(T, X, 1);

  ReadLn;

  LeaveGraphics;
end;

procedure PlotFFT(var Freq : array[LbF..UbF : Integer] of Float; 
                  var Z    : array[LbZ..UbZ : Integer] of Complex;  
                  Title    : String);
var
  Fq : array[0..MidIndex] of Float;   { Frequency }
  X  : array[0..MidIndex] of Float;   { Real part of FFT }
  Y  : array[0..MidIndex] of Float;   { Imag. part of FFT }

  Xmin, Xmax, Xstep       : Float;    { Ox scale }
  Yr_min, Yr_max, Yr_step : Float;    { Oy scale (real part) }
  Yi_min, Yi_max, Yi_step : Float;    { Oy scale (imag. part) }
  Ymin, Ymax, Ystep       : Float;    { Oy scale (global) }
  I                       : Integer;  { Loop variable }

begin
  if not InitGraphics(9, 2, '../../chr') then
    begin
      Writeln('Unable to set graphic mode!');
      Exit;
    end;

  SetWindow(15, 80, 15, 85, True);

  for I := 0 TO MidIndex do
    begin
      Fq[I] := Freq[I];
      X[I]  := Re(Z[I]);
      Y[I]  := Im(Z[I]);
    end;

  AutoScale(Fq, LinScale, Xmin, Xmax, Xstep);
  AutoScale(X,  LinScale, Yr_min, Yr_max, Yr_step);
  AutoScale(Y,  LinScale, Yi_min, Yi_max, Yi_step);

  Ymin  := Min(Yr_min, Yi_min);
  Ymax  := Max(Yr_max, Yi_max);
  Ystep := Min(Yr_step, Yi_step);

  SetOxScale(LinScale, Xmin, Xmax, Xstep);
  SetOyScale(LinScale, Ymin, Ymax, Ystep);

  SetOxTitle('Frequency (Hz)');
  SetOyTitle('FFT');
  SetGraphTitle(Title);

  SetPointParam(1, 0, 1, 1);  { Set point type to 0 so that }
  SetPointParam(2, 0, 1, 1);  { only lines will be plotted. }

  SetCurvLegend(1, 'Real');
  SetCurvLegend(2, 'Imag.');

  PlotOxAxis;
  PlotOyAxis;

  PlotGrid(BothGrid);

  WriteGraphTitle;

  WriteLegend(2, False, True);

  PlotCurve(Fq, X, 1);
  PlotCurve(Fq, Y, 2);

  ReadLn;

  LeaveGraphics;
end;

begin
  Assign(OutFile, 'testfft.txt');
  Rewrite(OutFile);

  for I := 0 to MaxIndex do
    begin
      T[I] := I * DT;
      Freq[I] := I * DF;
      InArray[I] := Cmplx(F(T[I]), 0.0);
    end;

  ListData(InArray, 'Time domain data before transform');
  PlotData(T, InArray, 'Original signal');

  FFT(InArray, OutArray);
  PlotFFT(Freq, OutArray, 'Fourier Transform');

  ListData(OutArray, 'Frequency domain data after transform');

  Close(OutFile);
end.
