{ ******************************************************************
  This program solves a system of linear equations, of the form
  AX = B, by the method of Gauss-Jordan. The method is demonstrated
  by solving a series of Hilbert systems of increasing order.
  Hilbert systems have ill-conditioned matrices (i.e. with
  determinants close to zero), so that the matrix is considered
  singular for an order which depends on the numerical precision
  of the software.

  The type of real numbers is defined by the compiler directives
  (see utypes.pas for details). The constant MachEp (defined in
  utypes.pas) sets the numerical precision which can be obtained
  with each type.
  ****************************************************************** }

program hilbert;

uses
  utypes, ugausjor, ustrings, matutils;

{$define Matrix array[Lb1..Ub1 : Integer; Lb2..Ub2 : Integer] of Float}

procedure HilbertSystem(var A : Matrix);
{ ------------------------------------------------------------------
  Generates the Hilbert system of order N

  A[1..N, 1..N] = system matrix :

        ( 1      1/2     1/3     1/4     ... 1/N      )
        ( 1/2    1/3     1/4     1/5     ... 1/(N+1)  )
    A = ( 1/3    1/4     1/5     1/6     ... 1/(N+2)  )
        ( ........................................... )
        ( 1/N    1/(N+1) 1/(N+2) 1/(N+3) ... 1/(2N-1) )

  A[1..N, M] = vector of constant terms (M = N + 1) :

               N
    A[i, M] = Sum A[i,j]
              j=1

  The solution vector is X = [1 1 1 ... 1]
  ------------------------------------------------------------------ }
var
  N, M : Integer;
  I, J : Integer;
  Sum  : Float;
begin
  N := Ub1;
  M := Ub2;

  { First row of matrix }
  A[1,1] := 1.0;
  for J := 2 to N do
    A[1,J] := 1.0 / J;

  for I := 2 to N do
    begin
      { N-th column of matrix }
      A[I,N] := 1.0 / (N + I - 1);
      { Fill matrix }
      for J := 1 to N - 1 do
        A[I,J] := A[I - 1,J + 1];
    end;

  { Last column = Constant vector }
  for I := 1 to N do
    begin
      Sum := 0.0;
      for J := 1 to N do
        Sum := Sum + A[I,J];
      A[I,M] := Sum;
    end;
end;

var
  N       : Integer;  { Order of the system }
  M       : Integer;  { N + 1 }
  ErrCode : Integer;  { Error code }
  D       : Float;    { Determinant }

begin
  N := 1;
  SetFormat(10, 2, True, True);

  repeat
    { Set system order }
    Inc(N);
    M := N + 1;

    { Allocate matrix }
    var A : array[1..N, 1..M] of Float;

    { Generate Hilbert system of order N }
    HilbertSystem(A);

    Writeln; Writeln('HILBERT SYSTEM OF ORDER ', N);
    if N < 7 then WriteMatrix('', A) else Writeln;

    { Solve Hilbert system }
    GaussJordan(A, D);
    ErrCode := MathErr;

    if ErrCode <> 0 then
      begin
        Writeln('Quasi-singular system');
        Halt;
    end;

    Writeln('Determinant = ', FloatStr(D));
    Writeln;

    if N < 7 then
      WriteMatrix('Inverse matrix and solution vector', A)
    else
      WriteMatrixCols('Solution vector', A, M, M);

    Writeln('Press ENTER to continue');
    Readln;

  until ErrCode <> 0;
end.
