{ ******************************************************************
  This program performs a nonlinear least squares fit of the
  Hill equation:

                                  Ymax . x^n
                              y = ----------
                                  K^n + x^n

  The parameters are: B(0) = Ymax, B(1) = K, B(2) = n

  The name of the input file (default extension = .DAT) is passed
  on the command line. The structure of this file is described in
  REGUTILS.PAS

  The file HILL.DAT is an example taken from a study in enzyme
  inhibition kinetics.
  ****************************************************************** }

program reghill;

uses
  utypes, uinterv, unlfit, uhillfit,
  uregtest, uinvbeta, uplot, regutils;

const
  MaxIter = 1000;    { Max. number of iterations }
  Tol     = 1.0E-4;  { Required precision }
  Alpha   = 0.05;    { Significance level }

var
  InFName : String;    { Name of input file }
  Title   : String;    { Title of study }
  XName,
  YName   : String;    { Names of variable }
  N       : Integer;   { Number of points }
  Test    : TRegTest;  { Regression tests }
  Sr      : Float;     { Residual standard dev. }
  Tc      : Float;     { Critical t value }
  Fc      : Float;     { Critical F value }
  I       : Integer;   { Loop variable }

  B : array[0..2] of Float;        { Regression parameters }
  V : array[0..2, 0..2] of Float;  { Variance-covariance matrix }

procedure ReadCmdLine(var InFName : String);
{ ------------------------------------------------------------------
  Reads command line
  ------------------------------------------------------------------ }
  begin
    { Name of input file }
    InFName := ParamStr(1);
    if InFName = '' then InFName := 'hill.dat';
    if Pos('.', InFName) = 0 then InFName := InFName + '.dat';
  end;

procedure SetBounds;
{ ------------------------------------------------------------------
  Initialize parameters and bounds
  ------------------------------------------------------------------ }
  var
    I : Integer;
  begin
    for I := 0 to 2 do
      B[I] := 0.0;

    { Ymax and K are assumed to be positive but n can be negative }

    SetParamBounds(0, 0, 1000);
    SetParamBounds(1, 0, 1000);
    SetParamBounds(2, -10, 10);
  end;

{ *************************** Main program ************************* }

begin
  { Read command line  }
  ReadCmdLine(InFName);

  { Read number of points }
  N := ReadNPoints(InFName, 4);
  if N < 0 then Halt;

  { Dimension arrays }
  var X, Y, S, Ycalc : array[1..N] of Float;

  { Read input file }
  if ReadInputFile(InFName, Title, XName, YName, X, Y) <> 0 then
    begin
      WriteLn('Error reading file ', InFName);
      Halt;
    end;

  { Set parameter bounds }
  SetBounds;

  { Perform regression }
  HillFit(X, Y, MaxIter, Tol, B, V);

  if MathErr <> MatOk then
    begin
      Writeln('Unable to fit curve! (MathErr = ', MathErr, ')');
      Halt;
    end;

  { Compute predicted Y values }
  for I := 1 to N do
    Ycalc[I] := HillFit_Func(X[I], B);

  { Update var-cov. matrix and compute statistical tests }
  RegTest(Y, Ycalc, V, Test);

  { Compute Student's t and Snedecor's F }
  Tc := InvStudent(Test.Nu2, 1 - 0.5 * Alpha);
  Fc := InvSnedecor(Test.Nu1, Test.Nu2, 1 - Alpha);

  { Update standard deviations }
  Sr := Sqrt(Test.Vr);
  for I := 1 to N do
    S[I] := Sr;

  { Write results }
  WriteResults(Title, N, X, Y, S, Ycalc, B, V, Test, Alpha, Tc, Fc);

  { Plot curve }
  PlotGraph(HillFit_Func, Title, X, Y, B);
end.
