{ ******************************************************************
  Cholesky factorization of a positive definite symmetric matrix
  ****************************************************************** }

unit ucholesk;

interface

uses
  utypes;

{$define Matrix array[Lb1..Ub1 : Integer; Lb2..Ub2 : Integer] of Float}

procedure Cholesky(var A, L : Matrix);
{ ------------------------------------------------------------------
  Cholesky decomposition. Factors the symmetric positive definite
  matrix A as a product L * L' where L is a lower triangular matrix.
  This procedure may be used as a test of positive definiteness.
  ------------------------------------------------------------------
  Possible results : MatOk     : No error
                     MatErrDim : Non-compatible dimensions
                     MatNotPD  : Matrix not positive definite
  ------------------------------------------------------------------ }

implementation

procedure Cholesky(var A, L : Matrix);
var
  I, J, K : Integer;
  Sum     : Float;
begin
  if (Lb1 <> Lb2) or (Ub1 <> Ub2) then
    begin
      SetErrCode(MatErrDim);
      Exit
    end;

  for K := Lb1 to Ub1 do
    begin
      Sum := A[K,K];
      for J := Lb2 to K - 1 do
        Sum := Sum - Sqr(L[K,J]);

      if Sum <= 0.0 then
        begin
          SetErrCode(MatNotPD);
          Exit
        end;

      L[K,K] := Sqrt(Sum);
      for I := K + 1 to Ub1 do
        begin
          Sum := A[I,K];
          for J := Lb2 to K - 1 do
            Sum := Sum - L[I,J] * L[K,J];
          L[I,K] := Sum / L[K,K];
          L[K,I] := 0.0;
        end;
    end;

  SetErrCode(MatOk);
end;

end.
