{ ******************************************************************
  Eigenvalues and eigenvectors of a symmetric matrix (SVD method)
  ****************************************************************** }

unit ueigsym;

interface

uses
  utypes, uswap, usvd;

{$define Vector array[Lb..Ub : Integer] of Float}

{$define Matrix array[Lb1..Ub1 : Integer; Lb2..Ub2 : Integer] of Float}

{$define VMatrix array[LbV1..UbV1 : Integer; LbV2..UbV2 : Integer] of Float}

procedure EigenSym(var A      : Matrix;
                   var Lambda : Vector;
                   var V      : VMatrix);
{ ------------------------------------------------------------------
  Eigenvalues and eigenvectors of a symmetric matrix by
  singular value decomposition
  ------------------------------------------------------------------
  Input parameters  : A      = matrix
  ------------------------------------------------------------------
  Output parameters : Lambda = eigenvalues in decreasing order
                      V      = matrix of eigenvectors (columns)
  ------------------------------------------------------------------
  Possible results  : MatOk
                      MatNonConv
  ------------------------------------------------------------------
  The eigenvectors are normalized, with their first component > 0
  This procedure destroys the original matrix A
  ------------------------------------------------------------------ }

implementation

procedure EigenSym(var A      : Matrix;
                   var Lambda : Vector;
                   var V      : VMatrix);
var
  I, J, K : Integer;
  R       : Float;
begin
  SV_Decomp(A, Lambda, V);

  if MathErr <> 0 then Exit;

  { Sort eigenvalues and eigenvectors }
  for I := Lb to Pred(Ub) do
    begin
      K := I;
      R := Lambda[I];
      for J := Succ(I) to Ub do
        if Lambda[J] > R then
          begin
            K := J;
            R := Lambda[J];
          end;

      FSwap(Lambda[I], Lambda[K]);
      for J := Lb to Ub do
        FSwap(V[J,I], V[J,K]);
    end;

  { Make sure that the first component of each eigenvector is > 0 }
  for J := Lb to Ub do
    if V[Lb,J] < 0.0 then
      for I := Lb to Ub do
        V[I,J] := - V[I,J];
end;

end.
