{ ******************************************************************
  This unit fits a sum of decreasing exponentials:
  y = Y0 + A1.exp(-a1.x) + A2.exp(-a2.x) + A3.exp(-a3.x) + ...
  ****************************************************************** }

unit uexpfit;

interface

uses
  utypes, umath, uswap, uqsort, umulfit, unlfit, urootpol;

{$i regdef.inc}

procedure ExpFit(var X, Y : Vector;
                 Nexp     : Integer;
                 ConsTerm : Boolean;
                 MaxIter  : Integer;
                 Tol      : Float;
                 var B    : BVector;
                 var V    : VMatrix);
{ ------------------------------------------------------------------
  Unweighted fit of sum of exponentials
  ------------------------------------------------------------------
  Input parameters:  X, Y     = point coordinates
                     Nexp     = number of exponentials
                     ConsTerm = presence of constant term B(0)
                     MaxIter  = max. number of iterations
                     Tol      = tolerance on parameters
  Output parameters: B        = regression parameters
                     V        = inverse matrix
  ------------------------------------------------------------------ }

procedure WExpFit(var X, Y, S : Vector;
                  Nexp        : Integer;
                  ConsTerm    : Boolean;
                  MaxIter     : Integer;
                  Tol         : Float;
                  var B       : BVector;
                  var V       : VMatrix);
{ ------------------------------------------------------------------
  Weighted fit of sum of exponentials
  ------------------------------------------------------------------
  Additional input parameter:
  S = standard deviations of observations
  ------------------------------------------------------------------ }

function ExpFit_Func(X : Float; var B : BVector) : Float;
{ ------------------------------------------------------------------
  Returns the value of the regression function at point X
  ------------------------------------------------------------------ }

implementation

var
  gNexp     : Integer;
  gConsTerm : Boolean;

function ExpFit_Func(X : Float; var B : BVector) : Float;
{ ------------------------------------------------------------------
  The function is of the form:

  y = Y0 + A1.exp(-a1.x) + A2.exp(-a2.x) + A3.exp(-a3.x) + ...

  with:

    B[0] = Y0
    B[1] = A1         B[2] = a1
    ...............................
    B[2*i-1] = Ai     B[2*i] = ai     i = 1..N_exp
  ------------------------------------------------------------------ }
  var
    I : Integer;
    S : Float;
  begin
    if gConsTerm then
      S := B[0]
    else
      S := 0.0;
    for I := 1 to gNexp do
      S := S + B[2 * I - 1] * Expo(- B[2 * I] * X);
    ExpFit_Func := S;
  end;

procedure ExpFit_Deriv(X, Y : Float; var B, D : BVector);
{ --------------------------------------------------------------------
  Computes the derivatives of the regression function at point (X, Y)
  with respect to the parameters B. The results are returned in D.
  D[I] contains the derivative with respect to the I-th parameter.
  -------------------------------------------------------------------- }
  var
    I, P, Q : Integer;
    E       : Float;
  begin
    if gConsTerm then
      D[0] := 1.0;                { dy/dY0 = 1 }
    for I := 1 to gNexp do
      begin
        Q := 2 * I;
        P := Pred(Q);
        E := Expo(- B[Q] * X);
        D[P] := E;                { dy/dAi = exp(-ai.x) }
        D[Q] := - X * B[P] * E;   { dy/dai = -x.Ai.exp(-ai.x) }
      end;
  end;

procedure SortPoints(var X, Y : Vector);
{ Sort points by increasing values of X }
var
  I, J, K : Integer;
  A       : Float;
begin
  for I := Lb to Pred(Ub) do
    begin
      K := I;
      A := X[I];
      for J := Succ(I) to Ub do
        if X[J] < A then
          begin
            K := J;
            A := X[J];
          end;

      FSwap(X[I], X[K]);
      FSwap(Y[I], Y[K]);
    end;
end;

procedure ApproxFit(Mode        : TRegMode;
                    var X, Y, S : Vector;
                    var B       : BVector);
{ --------------------------------------------------------------------
  Approximate fit of a sum of exponentials by linear regression
  --------------------------------------------------------------------
  Input :  Mode = OLS or WLS
           X, Y = point coordinates
           S    = standard deviations of Y

  Output : B    = estimated regression parameters
  --------------------------------------------------------------------
  Ref. : R. GOMENI & C. GOMENI, Automod : A polyalgorithm for an
         integrated analysis of linear pharmacokinetic models
         Comput. Biol. Med., 1979, 9, 39-48
  -------------------------------------------------------------------- }

  var
    I, K, M : Integer;
    Km1     : Integer;
    H       : Float;                   { Integration step }
    X1, Y1  : array[Lb..Ub] of Float;  { Modified coordinates }

  begin
    M := Pred(2 * gNexp);

    SortPoints(X, Y);

    for K := Lb to Ub do
      begin
        X1[K] := X[K];
        Y1[K] := Y[K];
      end;

    { Change scale so that the X's begin at zero }
    if X[Lb] <> 0.0 then
      for K := Lb to Ub do
        X1[K] := X1[K] - X[Lb];

    { Estimate the constant term Y0 then subtract it from each Y value }
    if gConsTerm then
      begin
        if Y[Lb] > Y[Ub] then   { Decreasing curve }
          B[0] := 0.9 * Y[Ub]
        else                    { Increasing curve }
          B[0] := 1.1 * Y[Ub];
        for K := Lb to Ub do
          Y1[K] := Y1[K] - B[0];
      end;

    { ----------------------------------------------------------------
      Fit the linearized form of the function :

      y = p(0) + p(1) * x + p(2) * x^2 + ... + p(Nexp-1) * x^(Nexp-1)

                   (x                         (x    (x
         + p(Nexp) | y dx + ... + p(2*Nexp-1) | ....| y dx
                   )0                         )0    )0
      ---------------------------------------------------------------- }

    var U : array[Lb..Ub, 1..M] of Float;  { Variables for linear fit }
    var P : array[0..M] of Float;          { Linear regression param. }
    var V : array[0..M, 0..M] of Float;    { Var-cov. matrix }

    { Compute increasing powers of X }
    if gNexp > 1 then
      for K := Succ(Lb) to Ub do
        begin
          U[K,1] := X1[K];
          for I := 2 to Pred(gNexp) do
            U[K,I] := U[K,I - 1] * X1[K];
        end;

    { Compute integrals by the trapezoidal rule }
    for K := Succ(Lb) to Ub do
      begin
        Km1 := K - 1;
        H := 0.5 * (X1[K] - X1[Km1]);
        U[K, gNexp] := U[Km1, gNexp] + (Y1[K] + Y1[Km1]) * H;
        for I := Succ(gNexp) to M do
          U[K,I] := U[Km1, I] + (U[K, I - 1] + U[Km1, I - 1]) * H;
      end;

    { Fit the equation }
    case Mode of
      OLS : MulFit(U, Y1, M, True, P, V);
      WLS : WMulFit(U, Y1, S, M, True, P, V);
    end;

    if MathErr <> MatOk then Exit;

    { ----------------------------------------------------------------
      The exponents are the roots of the polynomial :
      x^Nexp + p(Nexp) * x^(Nexp-1) - p(Nexp+1) * x^(Nexp-2) +...
      ---------------------------------------------------------------- }

    var C     : array[0..gNexp] of Float;    { Coeff. of polynomial }
    var Z     : array[1..gNexp] of Complex;  { Roots of polynomial }
    var Alpha : array[1..gNexp] of Float;    { Exponents }

    { Compute polynomial coefficients }
    C[gNexp] := 1.0;
    for I := 1 to gNexp do
      if Odd(I) then
        C[gNexp - I] := P[gNexp + I - 1]
      else
        C[gNexp - I] := - P[gNexp + I - 1];

    { Solve polynomial }
    if RootPol(C, Z) <> gNexp then
      begin
        SetErrCode(MatNonConv);
        Exit;
      end;

    { Get exponents and sort them in decreasing order }
    for I := 1 to gNexp do
      Alpha[I] := Re(Z[I]);
    DQSort(Alpha);

    { Compute the coefficients of the exponentials by
      linear regression on the exponential terms }

    var UU : array[Lb..Ub, 1..gNexp] of Float;    { Variables for linear fit }
    var PP : array[1..gNexp] of Float;            { Linear regression param. }
    var VV : array[1..gNexp, 1..gNexp] of Float;  { Var-cov. matrix }

    for K := Lb to Ub do
      for I := 1 to gNexp do
        UU[K,I] := Expo(- Alpha[I] * X1[K]);

    case Mode of
      OLS : MulFit(UU, Y1, gNexp, False, PP, VV);
      WLS : WMulFit(UU, Y1, S, gNexp, False, PP, VV);
    end;

    if MathErr <> MatOk then Exit;

    { Extract model parameters }
    for I := 1 to gNexp do
      begin
        { Correct for scale change if necessary }
        if X[1] <> 0.0 then
          PP[I] := PP[I] * Expo(Alpha[I] * X[1]);

        { Extract coefficients and exponents }
        B[2 * I - 1] := PP[I];  { Coefficients }
        B[2 * I] := Alpha[I];   { Exponents }
      end;
  end;

procedure GenExpFit(Mode        : TRegMode;
                    var X, Y, S : Vector;
                    Nexp        : Integer;
                    ConsTerm    : Boolean;
                    MaxIter     : Integer;
                    Tol         : Float;
                    var B       : BVector;
                    var V       : VMatrix);
begin
  gNexp := Nexp;
  gConsTerm := ConsTerm;

  if (GetOptAlgo in [NL_MARQ, NL_BFGS, NL_SIMP]) and NullParam(B) then
    ApproxFit(Mode, X, Y, S, B);

  if MaxIter = 0 then Exit;

  case Mode of
    OLS : NLFit(ExpFit_Func, ExpFit_Deriv, X, Y, MaxIter, Tol, B, V);
    WLS : WNLFit(ExpFit_Func, ExpFit_Deriv, X, Y, S, MaxIter, Tol, B, V);
  end;
end;

procedure ExpFit(var X, Y : Vector;
                 Nexp     : Integer;
                 ConsTerm : Boolean;
                 MaxIter  : Integer;
                 Tol      : Float;
                 var B    : BVector;
                 var V    : VMatrix);
var
  S : array[Lb..Ub] of Float;  { Dummy variable }
begin
  GenExpFit(OLS, X, Y, S, Nexp, ConsTerm, MaxIter, Tol, B, V);
end;

procedure WExpFit(var X, Y, S : Vector;
                  Nexp        : Integer;
                  ConsTerm    : Boolean;
                  MaxIter     : Integer;
                  Tol         : Float;
                  var B       : BVector;
                  var V       : VMatrix);
begin
  GenExpFit(WLS, X, Y, S, Nexp, ConsTerm, MaxIter, Tol, B, V);
end;

end.
