(*==========================================================================

    fourier.pas  -  Don Cross <dcross@intersrv.com>

    Modified by Jean Debord for use with GPMath.

    This is a Turbo Pascal Unit for calculating the Fast Fourier
    Transform (FFT) and the Inverse Fast Fourier Transform (IFFT).

    Revision history [most recent first]:

2007 December 10 [Jean Debord]
     Increased the theoretical number of points to about 2^26
     for a 32-bit compiler. The exact exponent is returned by
     function MaxPower.

2007 January 4 [Jean Debord]
     Modified for new TPMath version. Renamed as ufft.pas
     Now uses complex arrays.

1996 December 11 [Don Cross]
    Improved documentation of the procedure CalcFrequency.
    Fixed some messed up comments in procedure IFFT.

1996 December 6 [Don Cross]
    Made procedure 'fft_integer' more efficient when buffer size changes
    in successive calls:  the buffer is now only resized when the input
    has more samples, not a differing number of samples.
    Also changed the way 'fft_integer_cleanup' works so that it is
    more "bullet-proof".

1996 December 4 [Don Cross]
    Adding the procedure 'CalcFrequency', which calculates the FFT
    at a specific frequency index p=0..n-1, instead of the whole
    FFT.  This is O(n^2) instead of O(n*log(n)).

1996 November 30 [Don Cross]
    Adding a routine to allow FFT of an input array of integers.
    It is called 'fft_integer'.

1996 November 18 [Don Cross]
    Added some comments.

1996 November 17 [Don Cross]
    Wrote and debugged first version.

==========================================================================*)

unit ufft;

interface

uses
  utypes;

{$define IntVector array[LbI..UbI : Integer] of Integer}

{$define CompVector array[LbC..UbC : Integer] of Complex}

procedure FFT(var InArray, OutArray : CompVector);
{ -------------------------------------------------------------------
  Calculates the Fast Fourier Transform of the array of complex
  numbers represented by 'InArray' to produce the output complex
  numbers in 'OutArray'.
  ------------------------------------------------------------------- }

procedure IFFT(var InArray, OutArray : CompVector);
{ -------------------------------------------------------------------
  Calculates the Inverse Fast Fourier Transform of the array of
  complex numbers represented by 'InArray' to produce the output
  complex numbers in 'OutArray'.
  ------------------------------------------------------------------- }

procedure FFT_Integer(var RealIn,
                          ImagIn   : IntVector;
                      var OutArray : CompVector);
{ -------------------------------------------------------------------
  Same as procedure FFT, but uses Integer input arrays instead of
  double.  Make sure you call FFT_Integer_Cleanup after the last
  time you call FFT_Integer to free up memory it allocates.
  ------------------------------------------------------------------- }

procedure FFT_Integer_Cleanup;
{ -------------------------------------------------------------------
   If you call the procedure 'FFT_Integer', you must call
   'FFT_Integer_Cleanup' after the last time you call 'FFT_Integer'
   in order to free up dynamic memory.
  ------------------------------------------------------------------- }

procedure CalcFrequency(    FrequencyIndex : Integer;
                        var InArray        : CompVector;
                        var FT             : Complex);
{ -------------------------------------------------------------------
  This procedure calculates the complex frequency sample at a given
  index directly.  Use this instead of 'FFT' when you only need one
  or two frequency samples, not the whole spectrum.

  It is also useful for calculating the Discrete Fourier Transform (DFT)
  of a number of data which is not an integer power of 2. For example,
  you could calculate the DFT of 100 points instead of rounding up to
  128 and padding the extra 28 array slots with zeroes.
  ------------------------------------------------------------------- }

implementation

const
  MaxPower = Trunc(Ln(MaxInt / SizeOf(Complex)) / Ln(2.0));

type
  TTempArray(Ub : Integer) = array[0..Ub] of Complex;

var
  TempArraySize : Integer = 0;  { Flag that buffer Temp is not allocated }
  Temp          : ^TTempArray;

function IsPowerOfTwo(X : Integer) : Boolean;
var
  I, Y : Integer;
begin
  Y := 2;
  for I := 1 to MaxPower do
    begin
      if X = Y then
        begin
          IsPowerOfTwo := True;
          Exit;
        end;
      Y := Y shl 1;
    end;
  IsPowerOfTwo := False;
end;

function NumberOfBitsNeeded(PowerOfTwo : Integer) : Integer;
var
  I : Integer;
begin
  for I := 0 to MaxPower do
    begin
      if (PowerOfTwo and (1 shl I)) <> 0 then
        begin
          NumberOfBitsNeeded := I;
          Exit;
        end;
    end;
end;

function ReverseBits(Index, NumBits : Integer) : Integer;
var
  I, Rev : Integer;
begin
  Rev := 0;
  for I := 0 to NumBits - 1 do
    begin
      Rev := (Rev shl 1) or (Index and 1);
      Index := Index shr 1;
    end;
  ReverseBits := Rev;
end;

procedure FourierTransform(AngleNumerator : Float;
                           var InArray,
                               OutArray   : CompVector);
var
  NumSamples, NumBits   : Integer;
  I, J, K, N            : Integer;
  BlockSize, BlockEnd   : Integer;
  Delta_angle, Delta_ar : Float;
  Alpha, Beta           : Float;
  A, T                  : Complex;

begin
  NumSamples := UbC + 1;

  if (LbC <> 0) or (UbC < 1) or (not IsPowerOfTwo(NumSamples)) then
    begin
      SetErrCode(-1);
      Exit;
    end;

  SetErrCode(0);
  NumBits := NumberOfBitsNeeded(NumSamples);
  for I := LbC to UbC do
    begin
      J := ReverseBits(I, NumBits);
      OutArray[J] := InArray[I];
    end;

  BlockEnd := 1;
  BlockSize := 2;

  while BlockSize <= NumSamples do
    begin
      Delta_angle := AngleNumerator / BlockSize;
      Alpha := Sin(0.5 * Delta_angle);
      Alpha := 2.0 * Alpha * Alpha;
      Beta := Sin(Delta_angle);

      I := 0;
      while I < NumSamples do
        begin
          A := Cmplx(1.0, 0.0);
          J := I;
          for N := 0 to BlockEnd - 1 do
            begin
              K := J + BlockEnd;
              T := A * OutArray[K];
              OutArray[K] := OutArray[J] - T;
              OutArray[J] := OutArray[J] + T;
              Delta_ar := Alpha * Re(A) + Beta * Im(A);
              A := Cmplx(Re(A) - Delta_ar,
                         Im(A) - (Alpha * Im(A) - Beta * Re(A)));
              Inc(J);
            end;
          I := I + BlockSize;
        end;

      BlockEnd := BlockSize;
      BlockSize := BlockSize shl 1;
    end;
end;

procedure FFT(var InArray, OutArray : CompVector);
begin
  FourierTransform(2 * Pi, InArray, OutArray);
end;

procedure IFFT(var InArray, OutArray : CompVector);
var
  NumSamples, I : Integer;
begin
  FourierTransform(- 2 * Pi, InArray, OutArray);
  if MathErr <> 0 then Exit;

  { Normalize the resulting time samples }
  NumSamples := UbC + 1;
  for I := LbC to UbC do
    OutArray[I] := OutArray[I] / NumSamples;
end;

procedure FFT_Integer(var RealIn,
                          ImagIn   : IntVector;
                      var OutArray : CompVector);
var
  I : Integer;
begin
  if (LbI <> 0) or (LbI <> LbC) or (UbI <> UbC) then
    begin
      SetErrCode(-1);
      Exit;
    end;

  { Free up memory in case we already have some }
  if UbC > TempArraySize then
    begin
      FFT_Integer_Cleanup;
      New(Temp, UbC);
      TempArraySize := UbC;
    end;

  for I := LbC to UbC do
    Temp^[I] := Cmplx(RealIn[I], ImagIn[I]);

  FourierTransform(2 * Pi, Temp^, OutArray);
end;

procedure FFT_Integer_Cleanup;
begin
  if TempArraySize > 0 then
    begin
      TempArraySize := 0;
      Dispose(Temp);
    end;
end;

procedure CalcFrequency(    FrequencyIndex : Integer;
                        var InArray        : CompVector;
                        var FT             : Complex);
var
  NumSamples, K    : Integer;
  Cos1, Cos2, Cos3 : Float;
  Sin1, Sin2, Sin3 : Float;
  Theta, Beta      : Float;

begin
  NumSamples := UbC + 1;
  FT := Cmplx(0.0, 0.0);

  Theta := 2 * Pi * FrequencyIndex / NumSamples;
  Sin1 := Sin(- 2 * Theta);
  Sin2 := Sin(- Theta);
  Cos1 := Cos(- 2 * Theta);
  Cos2 := Cos(- Theta);
  Beta := 2 * Cos2;

  for K := LbC to UbC do
    begin
      { Update trig values }
      Sin3 := Beta * Sin2 - Sin1;
      Sin1 := Sin2;
      Sin2 := Sin3;

      Cos3 := Beta * Cos2 - Cos1;
      Cos1 := Cos2;
      Cos2 := Cos3;

      FT := FT + InArray[K] * Cmplx(Cos3, Sin3);
    end;
end;

end.
