{ ******************************************************************
  This unit fits a rational fraction :

                           p0 + p1.x + p2.x^2 + ...
                       y = ------------------------
                           1 + q1.x + q2.x^2 + ...

  ****************************************************************** }

unit ufracfit;

interface

uses
  utypes, upolynom, umulfit, unlfit;

{$i regdef.inc}

procedure FracFit(var X, Y   : Vector;
                  Deg1, Deg2 : Integer;
                  ConsTerm   : Boolean;
                  MaxIter    : Integer;
                  Tol        : Float;
                  var B      : BVector;
                  var V      : VMatrix);
{ ------------------------------------------------------------------
  Unweighted fit of rational fraction
  ------------------------------------------------------------------
  Input parameters:  X, Y       = point coordinates
                     Deg1, Deg2 = degrees of numerator and denominator
                     ConsTerm   = presence of constant term p0
                     MaxIter    = max. number of iterations
                     Tol        = tolerance on parameters
  Output parameters: B          = regression parameters
                     V          = inverse matrix
  ------------------------------------------------------------------ }

procedure WFracFit(var X, Y, S : Vector;
                   Deg1, Deg2  : Integer;
                   ConsTerm    : Boolean;
                   MaxIter     : Integer;
                   Tol         : Float;
                   var B       : BVector;
                   var V       : VMatrix);
{ ------------------------------------------------------------------
  Weighted fit of rational fraction
  ------------------------------------------------------------------
  Additional input parameter:
  S = standard deviations of observations
  ------------------------------------------------------------------ }

function FracFit_Func(X : Float; var B : BVector) : Float;
{ ------------------------------------------------------------------
  Returns the value of the regression function at point X
  ------------------------------------------------------------------ }


implementation

var
  gDeg1     : Integer = 1;     { Degree of numerator }
  gDeg2     : Integer = 1;     { Degree of denominator }
  gConsTerm : Boolean = True;  { Flags the presence of a constant term p0 }

  function FracFit_Func(X : Float; var B : BVector) : Float;
  { ----------------------------------------------------------------
    Computes the regression function at point X
    B is the vector of parameters, such that :

    B[0] = p0
    B[1] = p1            B[2] = p2         ...

    B[Deg1 + 1] = q1     B[Deg1 + 2] = q2  ...
    ---------------------------------------------------------------- }
  begin
    FracFit_Func := RFrac(X, B, gDeg1, gDeg2);
  end;

  procedure FracFit_Deriv(X, Y : Float; var B, D : BVector);
  { ----------------------------------------------------------------
    Computes the derivatives of the regression function at point (X,Y)
    with respect to the parameters B. The results are returned in D.
    D[I] contains the derivative with respect to the I-th parameter
    ---------------------------------------------------------------- }
  var
    I   : Integer;
    Den : Float;
  begin
    { Compute denominator (1 + q1.x + q2.x^2 + ...) }
    Den := 0.0;
    for I := (gDeg1 + gDeg2) downto Succ(gDeg1) do
      Den := (Den + B[I]) * X;
    Den := 1.0 + Den;

    { dy/dp0 = 1 / (1 + q1.x + q2.x^2 + ...) }
    if gConsTerm then D[0] := 1.0 / Den;

    { dy/dpi = x^i / (1 + q1.x + q2.x^2 + ...) }
    D[1] := X / Den;
    for I := 2 to gDeg1 do
      D[I] := D[I - 1] * X;

    { dy/dq1 = -x.y / (1 + q1.x + q2.x^2 + ...) }
    D[gDeg1 + 1] := - X * Y / Den;

    { dy/dqi = -x^i.y / (1 + q1.x + q2.x^2 + ...) }
    for I := (gDeg1 + 2) to (gDeg1 + gDeg2) do
      D[I] := D[I - 1] * X;
  end;

  procedure ApproxFit(Mode        : TRegMode;
                      var X, Y, S : Vector;
                      var B       : BVector);
  { ----------------------------------------------------------------
    Approximate fit of a rational fraction by linear regression:
    y = p0 + p1.x + p2.x^2 + ... - q1.(x.y) - q2.(x^2.y) - ...
    ----------------------------------------------------------------
    Input :  Mode = OLS or WLS
             X, Y = point coordinates
             S    = standard deviations
    Output : B    = estimated regression parameters
    ---------------------------------------------------------------- }
  var
    I, J : Integer;
  begin
    var U : array[Lb..Ub, 1..UbB] of Float;
    var V : array[LbB..UbB, LbB..UbB] of Float;

    for I := Lb to Ub do
      begin
        U[I,1] := X[I];
        for J := 2 to gDeg1 do
          U[I,J] := U[I, J - 1] * X[I];
        U[I, gDeg1 + 1] := - X[I] * Y[I];
        for J := (gDeg1 + 2) to UbB do
          U[I,J] := U[I, J - 1] * X[I];
      end;

    case Mode of
      OLS : MulFit(U, Y, UbB, gConsTerm, B, V);
      WLS : WMulFit(U, Y, S, UbB, gConsTerm, B, V);
    end;
  end;

  procedure GenFracFit(Mode        : TRegMode;
                       var X, Y, S : Vector;
                       Deg1, Deg2  : Integer;
                       ConsTerm    : Boolean;
                       MaxIter     : Integer;
                       Tol         : Float;
                       var B       : BVector;
                       var V       : VMatrix);
  begin
    gDeg1 := Deg1;
    gDeg2 := Deg2;
    gConsTerm := ConsTerm;

    if (GetOptAlgo in [NL_MARQ, NL_BFGS, NL_SIMP]) and NullParam(B) then
      ApproxFit(Mode, X, Y, S, B);

    if MaxIter = 0 then Exit;

    case Mode of
      OLS : NLFit(FracFit_Func, FracFit_Deriv, X, Y, MaxIter, Tol, B, V);
      WLS : WNLFit(FracFit_Func, FracFit_Deriv, X, Y, S, MaxIter, Tol, B, V);
    end;
  end;

  procedure FracFit(var X, Y   : Vector;
                    Deg1, Deg2 : Integer;
                    ConsTerm   : Boolean;
                    MaxIter    : Integer;
                    Tol        : Float;
                    var B      : BVector;
                    var V      : VMatrix);
  var
    S : array[Lb..Ub] of Float;  { Dummy variable }
  begin
    GenFracFit(OLS, X, Y, S, Deg1, Deg2, ConsTerm, MaxIter, Tol, B, V);
  end;

  procedure WFracFit(var X, Y, S : Vector;
                     Deg1, Deg2  : Integer;
                     ConsTerm    : Boolean;
                     MaxIter     : Integer;
                     Tol         : Float;
                     var B       : BVector;
                     var V       : VMatrix);
  begin
    GenFracFit(WLS, X, Y, S, Deg1, Deg2, ConsTerm, MaxIter, Tol, B, V);
  end;

end.
