{ ******************************************************************
  This unit fits the gamma variate regression model:

                y = a * (x - b)^c * exp[ - (x - b) / d]

  ****************************************************************** }

unit ugamfit;

interface

uses
  utypes, umath, umulfit, unlfit;

{$i regdef.inc}

procedure GammaFit(var X, Y : Vector;
                   MaxIter  : Integer;
                   Tol      : Float;
                   var B    : BVector;
                   var V    : VMatrix);
{ ------------------------------------------------------------------
  Unweighted fit of model
  ------------------------------------------------------------------
  Input parameters:  X, Y    = point coordinates
                     MaxIter = max. number of iterations
                     Tol     = tolerance on parameters
  Output parameters: B       = regression parameters
                     V       = inverse matrix
  ------------------------------------------------------------------ }

procedure WGammaFit(var X, Y, S : Vector;
                    MaxIter     : Integer;
                    Tol         : Float;
                    var B       : BVector;
                    var V       : VMatrix);
{ ------------------------------------------------------------------
  Weighted fit of model
  ------------------------------------------------------------------
  Additional input parameter:
  S = standard deviations of observations
  ------------------------------------------------------------------ }

function GammaFit_Func(X : Float; var B : BVector) : Float;
{ ------------------------------------------------------------------
  Returns the value of the regression function at point X
  ------------------------------------------------------------------ }

implementation

function GammaFit_Func(X : Float; var B : BVector) : Float;
{ ------------------------------------------------------------------
  Computes the regression function at point X
  B is the vector of parameters, such that :

  B[1] = a     B[2] = b     B[3] = c     B[4] = d
  ------------------------------------------------------------------ }
var
  W, Z : Float;
begin
  if X <= B[2] then
    begin
      GammaFit_Func := 0.0;
      Exit;
    end;

  W := X - B[2];                    { x - b }
  Z := B[3] * Ln(W) - W / B[4];     { c * ln(x - b) - (x - b) / d }

  if Z > MinLog then
    GammaFit_Func := B[1] * Exp(Z)  { a * (x - b)^c * exp[ - (x - b) / d] }
  else
    GammaFit_Func := 0.0;
end;

procedure GammaFit_Deriv(X, Y : Float; var B, D : BVector);
{ ------------------------------------------------------------------
  Computes the derivatives of the regression function at point X
  with respect to the parameters B. The results are returned in D.
  D[I] contains the derivative with respect to the I-th parameter.
  ------------------------------------------------------------------ }
var
  W : Float;
  I : Integer;
begin
  if X <= B[2] then
    begin
      for I := 1 to 4 do D[I] := 0.0;
      Exit;
    end;

  W := X - B[2];                        { w = x - b }
  D[1] := Y / B[1];                     { dy/da = y / a }
  D[2] := Y * (1.0 / B[4] - B[3] / W);  { dy/db = y * (1 / d - c / w) }
  D[3] := Y * Ln(W);                    { dy/dc = y * Ln(w) }
  D[4] := Y * W / Sqr(B[4]);            { dy/dd = y * w / d^2 }
end;

procedure ApproxFit(Mode        : TRegMode;
                    var X, Y, S : Vector;
                    var B       : BVector);
{ ------------------------------------------------------------------
  Computes initial estimates of the regression parameters on the
  linearized form of the function:

  ln y = ln a + c * ln w - w / d    w = x - b
  ------------------------------------------------------------------
  Input :  Mode = OLS or WLS
           X, Y = point coordinates
  Output : B    = estimated regression parameters
  ------------------------------------------------------------------ }
var
  K : Integer;  { Loop variable }
  P : Integer;  { Nb of points for linear fit }
  W : Float;    { x - b }

  A : array[0..2] of Float;        { Linear regression param. }
  V : array[0..2, 0..2] of Float;  { Variance-covariance matrix }

begin
  { Count number of points with Y > 0 }
  P := 0;
  for K := Lb to Ub do
    if Y[K] > 0.0 then
      Inc(P);

  if P < 4 then
    begin
      SetErrCode(MatSing);
      Exit;
    end;

  { Find the lowest X having a positive Y }
  W := MaxNum;
  for K := Lb to Ub do
    if (X[K] > 0.0) and (Y[K] > 0.0) and (X[K] < W) then
      W := X[K];

  { Set b slightly below this X value }
  B[2] := 0.9 * W;

  var U      : array[1..P, 1..2] of Float;  { Variables for linear fit }
  var Y1, S1 : array[1..P] of Float;        { ln y and standard dev. }

  P := 0;
  for K := Lb to Ub do
    if Y[K] > 0.0 then
      begin
        Inc(P);
        W := X[K] - B[2];
        U[P,1] := Ln(W);
        U[P,2] := W;
        Y1[P] := Ln(Y[K]);
        S1[P] := 1.0 / Y[K];
        if Mode = WLS then S1[P] := S1[P] * S[K];
      end;

  WMulFit(U, Y1, S1, 2, True, A, V);

  if MathErr = MatOk then
    begin
      B[1] := Exp(A[0]);
      B[3] := A[1];
      B[4] := - 1.0 / A[2];
    end;
end;

procedure GenGammaFit(Mode        : TRegMode;
                      var X, Y, S : Vector;
                      MaxIter     : Integer;
                      Tol         : Float;
                      var B       : BVector;
                      var V       : VMatrix);
begin
  if (GetOptAlgo in [NL_MARQ, NL_BFGS, NL_SIMP]) and NullParam(B) then
    ApproxFit(Mode, X, Y, S, B);

  if MaxIter = 0 then Exit;

  case Mode of
    OLS : NLFit(GammaFit_Func, GammaFit_Deriv, X, Y, MaxIter, Tol, B, V);
    WLS : WNLFit(GammaFit_Func, GammaFit_Deriv, X, Y, S, MaxIter, Tol, B, V);
  end;
end;

procedure GammaFit(var X, Y : Vector;
                   MaxIter  : Integer;
                   Tol      : Float;
                   var B    : BVector;
                   var V    : VMatrix);
var
  S : array[Lb..Ub] of Float;  { Dummy variable }
begin
  GenGammaFit(OLS, X, Y, S, MaxIter, Tol, B, V);
end;

procedure WGammaFit(var X, Y, S : Vector;
                    MaxIter     : Integer;
                    Tol         : Float;
                    var B       : BVector;
                    var V       : VMatrix);
begin
  GenGammaFit(WLS, X, Y, S, MaxIter, Tol, B, V);
end;

end.
