{ ******************************************************************
  This unit fits the Integrated Michaelis-Menten equation:

  y = s0 - Km * W(s0 / Km * exp((s0 - kcat * e0 * t) / Km))

  y    = product concentration at time t
  s0   = initial substrate concentration
  Km   = Michaelis constant
  kcat = catalytic constant
  e0   = total enzyme concentration

  W is Lambert's function (reciprocal of x * exp(x))

  The independent variable x may be:

  *  t  ==> fitted parameters : s0 (optional), Km, Vmax = kcat * e0
  *  s0 ==> fitted parameters : Km, (Vmax * t)
  *  e0 ==> fitted parameters : s0 (optional), Km, (kcat * t),
  ****************************************************************** }

unit umintfit;

interface

uses
  utypes, umath, ulambert, umeansd, ulinfit, unlfit;

{$i regdef.inc}

procedure MintFit(var X, Y : Vector;
                  MintVar  : TMintVar;
                  S0       : Float;
                  MaxIter  : Integer;
                  Tol      : Float;
                  var B    : BVector;
                  var V    : VMatrix);
{ ------------------------------------------------------------------
  Unweighted fit of model
  ------------------------------------------------------------------
  Input parameters:  X, Y    = point coordinates
                     MintVar = independant var. (Var_T, Var_S, Var_E)
                     S0      = initial substrate conc. (<=0 if it must
                               be fitted, for Var_T or Var_E only)
                     MaxIter = max. number of iterations
                     Tol     = tolerance on parameters
  Output parameters: B       = regression parameters
                     V       = inverse matrix
  ------------------------------------------------------------------ }

procedure WMintFit(var X, Y, S : Vector;
                   MintVar     : TMintVar;
                   S0          : Float;
                   MaxIter     : Integer;
                   Tol         : Float;
                   var B       : BVector;
                   var V       : VMatrix);
{ ------------------------------------------------------------------
  Weighted fit of model
  ------------------------------------------------------------------
  Additional input parameter:
  S = standard deviations of observations
  ------------------------------------------------------------------ }

function MintFit_Func(X : Float; var B : BVector) : Float;
{ ------------------------------------------------------------------
  Returns the value of the regression function at point X
  ------------------------------------------------------------------ }

implementation

var
  gMintVar : TMintVar;
  gFit_S0  : Boolean;
  gS0      : Float;

function MintFit_Func(X : Float; var B : BVector) : Float;
{ ------------------------------------------------------------------
  Computes the regression function at point X
  ------------------------------------------------------------------ }
begin
  if gMintVar = Var_S then
    MintFit_Func :=
      X - B[1] *
        LambertW(X / B[1] * Expo((X - B[2]) / B[1]),
                 True, False)
  else
    begin
      if gFit_S0 then gS0 := B[0];
      MintFit_Func := gS0 - B[1] *
        LambertW(gS0 / B[1] * Expo((gS0 - B[2] * X) / B[1]),
                 True, False);
    end;
end;

procedure MintFit_Deriv(X, Y : Float; var B, D : BVector);
{ ------------------------------------------------------------------
  Computes the derivatives of the regression function at point X
  with respect to the parameters B. The results are returned in D.
  D[I] contains the derivative with respect to the I-th parameter.
  ------------------------------------------------------------------ }
var
  L, Q, R : Float;
begin
  if gMintVar = Var_S then
    begin
      L := (X - Y) / B[1];
      D[2] := L / (1.0 + L);
      D[1] := D[2] * (Y - B[2]) / B[1];
    end
  else
    begin
      if gFit_S0 then gS0 := B[0];
      L := (gS0 - Y) / B[1];
      Q := 1.0 / (1.0 + L);
      R := L * Q;
      if gFit_S0 then D[0] := Y * Q / gS0;
      D[1] := R * (Y - B[2] * X) / B[1];
      D[2] := R * X;
    end;
end;

procedure ApproxFit(var X, Y : Vector; var B : BVector);
{ ------------------------------------------------------------------
  Computes initial estimates of the regression parameters by linear
  regression:

  x = t  ==> y / x = Vmax       + Km (1 / x) Ln(1 - y / s0)
  x = s0 ==> y     = (Vmax * t) + Km Ln(1 - y / s0)
  x = e0 ==> y / x = (kcat * t) + Km (1 / x) Ln(1 - y / s0)
  ------------------------------------------------------------------
  Input :  X, Y = point coordinates
  Output : B    = estimated regression parameters
  ------------------------------------------------------------------ }
  var
    I      : Integer;                     { Loop variable }
    XX, YY : array[Lb..Ub] of Float;      { Transformed coordinates }
    A      : array[0..1] of Float;        { Linear regression param. }
    V      : array[0..1, 0..1] of Float;  { Variance-covariance matrix }
  begin
    { Compute transformed coordinates }
    if gMintVar = Var_S then
      for I := Lb to Ub do
        begin
          XX[I] := Ln(1.0 - Y[I] / X[I]);
          YY[I] := Y[I];
        end
    else
      begin
        if gFit_S0 then
          begin
            B[0] := 1.1 * MaxVal(Y);
            gS0 := B[0];
          end;
        for I := Lb to Ub do
          begin
            XX[I] := Ln(1.0 - Y[I] / gS0) / X[I];
            YY[I] := Y[I] / X[I];
          end;
      end;

    { Perform linear regression }
    LinFit(XX, YY, A, V);

    { Retrieve parameters }
    B[1] := A[1];          { Km }
    B[2] := A[0];          { Vmax or (Vmax * t) or (kcat * t) }
  end;

procedure GenMintFit(Mode        : TRegMode;
                     var X, Y, S : Vector;
                     MintVar     : TMintVar;
                     S0          : Float;
                     MaxIter     : Integer;
                     Tol         : Float;
                     var B       : BVector;
                     var V       : VMatrix);
begin
  gMintVar := MintVar;
  gFit_S0 := (S0 <= 0.0);
  gS0 := S0;

  if (GetOptAlgo in [NL_MARQ, NL_BFGS, NL_SIMP]) and NullParam(B) then
    ApproxFit(X, Y, B);

  if MaxIter = 0 then Exit;

  case Mode of
    OLS : NLFit(MintFit_Func, MintFit_Deriv, X, Y, MaxIter, Tol, B, V);
    WLS : WNLFit(MintFit_Func, MintFit_Deriv, X, Y, S, MaxIter, Tol, B, V);
  end;
end;

procedure MintFit(var X, Y : Vector;
                  MintVar  : TMintVar;
                  S0       : Float;
                  MaxIter  : Integer;
                  Tol      : Float;
                  var B    : BVector;
                  var V    : VMatrix);
var
  S : array[Lb..Ub] of Float;  { Dummy variable }
begin
  GenMintFit(OLS, X, Y, S, MintVar, S0, MaxIter, Tol, B, V);
end;

procedure WMintFit(var X, Y, S : Vector;
                   MintVar     : TMintVar;
                   S0          : Float;
                   MaxIter     : Integer;
                   Tol         : Float;
                   var B       : BVector;
                   var V       : VMatrix);
begin
  GenMintFit(WLS, X, Y, S, MintVar, S0, MaxIter, Tol, B, V);
end;

end.
