{ ******************************************************************
  Minimization of a function of several variables by the
  Newton-Raphson method
  ****************************************************************** }

unit unewton;

interface

uses
  utypes, ulineq, ulinmin, ucompvec;

{$define Vector array[Lb..Ub : Integer] of Float}

{$define GradVector array[LbG..UbG : Integer] of Float}

{$define Matrix array[Lb1..Ub1 : Integer; Lb2..Ub2 : Integer] of Float}

procedure SaveNewton(FileName : string);
{ ------------------------------------------------------------------
  Save Newton-Raphson iterations in a file
  ------------------------------------------------------------------ }

procedure Newton(    Func     : TFuncNVar;
                     HessGrad : THessGrad;
                 var X        : Vector;
                     MaxIter  : Integer;
                     Tol      : Float;
                 var F_min    : Float;
                 var G        : GradVector;
                 var H_inv    : Matrix;
                 var Det      : Float);
{ ------------------------------------------------------------------
  Minimization of a function of several variables by the
  Newton-Raphson method
  ------------------------------------------------------------------
  Input parameters  : Func     = objective function
                      HessGrad = procedure to compute hessian and gradient
                      X        = initial minimum coordinates
                      MaxIter  = maximum number of iterations
                      Tol      = required precision
  ------------------------------------------------------------------
  Output parameters : X        = refined minimum coordinates
                      F_min    = function value at minimum
                      G        = gradient vector
                      H_inv    = inverse hessian matrix
                      Det      = determinant of hessian
  ------------------------------------------------------------------
  Possible results  : OptOk      = no error
                      OptNonConv = non-convergence
                      OptSing    = singular hessian matrix
                      MatErrDim  = non-compatible dimensions
  ---------------------------------------------------------------------- }

implementation

var
  WriteLogFile : Boolean = False;
  LogFile      : Text;

procedure SaveNewton(FileName : string);
  begin
    Assign(LogFile, FileName);
    Rewrite(LogFile);
    WriteLogFile := True;
  end;

procedure Newton(    Func     : TFuncNVar;
                     HessGrad : THessGrad;
                 var X        : Vector;
                     MaxIter  : Integer;
                     Tol      : Float;
                 var F_min    : Float;
                 var G        : GradVector;
                 var H_inv    : Matrix;
                 var Det      : Float);

  var
    I, Iter      : Integer;
    R            : Float;
    OldX, DeltaX : array[Lb..Ub] of Float;

  procedure Init;
  { Initializes variables }
  var
    I : Integer;
  begin
    { Initialize function }
    F_min := Func(X);

    { Initialize gradient and hessian (stored in H_inv) }
    HessGrad(X, G, H_inv);

    { Initialize search direction }
    for I := Lb to Ub do
      DeltaX[I] := - G[I];

    { Solve system }
    LinEq(H_inv, DeltaX, Det);
  end;

  procedure Terminate(ErrCode : Integer);
  { Set error code and deallocate arrays }
  begin
    SetErrCode(ErrCode);
    if WriteLogFile then
      Close(LogFile);
  end;

  begin
    if (Lb <> Lb1) or (Lb <> Lb2) or (Lb <> LbG) or
       (Ub <> Ub1) or (Ub <> Ub2) or (Ub <> UbG) then
         begin
           SetErrCode(MatErrDim);
           Exit;
         end;

    Init;

    if MathErr <> MatOk then
      begin
        Terminate(OptSing);
        Exit;
      end;

    if MaxIter < 1 then
      begin
        Terminate(OptOk);
        Exit;
      end;

    if WriteLogFile then
      begin
        WriteLn(LogFile, 'Newton-Raphson');
        WriteLn(LogFile, 'Iter         F');
      end;

    Iter := 0;

    repeat
      { Prepare next iteration }
      if WriteLogFile then
        WriteLn(LogFile, Iter:4, '   ', F_min:12);

      Iter := Iter + 1;
      if Iter > MaxIter then
        begin
          Terminate(OptNonConv);
          Exit;
        end;

      { Save old parameters }
      for I := Lb to Ub do
        OldX[I] := X[I];

      { Minimize along the direction specified by DeltaX }
      R := 0.1;
      LinMin(Func, X, DeltaX, R, 10, 0.01, F_min);

      { Compute new gradient and hessian }
      HessGrad(X, G, H_inv);

      { Initialize search direction }
      for I := Lb to Ub do
        DeltaX[I] := - G[I];

      { Solve system }
      LinEq(H_inv, DeltaX, Det);

      if MathErr <> MatOk then
        begin
          Terminate(OptSing);
          Exit;
        end;
    until CompVec(X, OldX, Tol);

    Terminate(OptOk);
  end;

end.

