{ ******************************************************************
  Multiple linear regression (Singular Value Decomposition)
  ****************************************************************** }

unit usvdfit;

interface

uses
  utypes, usvd;

{$i regdef.inc}

procedure SVDFit(var X    : Matrix;
                 var Y    : Vector;
                 ConsTerm : Boolean;
                 SVDTol   : Float;
                 var B    : BVector;
                 var V    : VMatrix);
{ ------------------------------------------------------------------
  Multiple linear regression: Y = B(0) + B(1) * X + B(2) * X2 + ...
  ------------------------------------------------------------------
  Input parameters:  X        = matrix of independent variables
                     Y        = vector of dependent variable
                     ConsTerm = presence of constant term B(0)
                     SVDTol   = tolerance on singular values
  Output parameters: B        = regression parameters
                     V        = inverse matrix
  ------------------------------------------------------------------ }

procedure WSVDFit(var X    : Matrix;
                  var Y, S : Vector;
                  ConsTerm : Boolean;
                  SVDTol   : Float;
                  var B    : BVector;
                  var V    : VMatrix);
{ ------------------------------------------------------------------
  Weighted multiple linear regression
  ------------------------------------------------------------------
  S = standard deviations of observations
  Other parameters as in SVDFit
  ------------------------------------------------------------------ }

implementation

procedure GenSVDFit(Mode     : TRegMode;
                    var X    : Matrix;
                    var Y, S : Vector;
                    ConsTerm : Boolean;
                    SVDTol   : Float;
                    var B    : BVector;
                    var V    : VMatrix);
{ ------------------------------------------------------------------
  General multiple linear regression routine (SVD algorithm)
  ------------------------------------------------------------------ }
var
  N, L, M : Integer;  { Array bounds }
  I, J, K : Integer;  { Loop variables }
  I1, J1  : Integer;  { Loop variables }
  Sigma   : Float;    { Square root of weight }

begin
  { Prepare arrays for SVD : U[L..N, L..M] and Z[L..N]
    since the lower bounds of U for the SVD routine must be
    the same in both dimensions }

  L := 0;
  N := Ub1 - Lb1;
  M := Ub2;

  if not ConsTerm then
    begin
      Inc(L);
      Inc(N);
    end;

  { Check dimensions }
  if (Lb <> Lb1) or (Ub <> Ub1) or
     (LbB <> L)  or (LbV1 <> L) or (LbV2 <> L) or
     (UbB <> M)  or (UbV1 <> M) or (UbV2 <> M) then
      begin
        SetErrCode(MatErrDim);
        Exit;
      end;

  { Check standard deviations }
  if Mode = WLS then
    for K := Lb to Ub do
      if S[K] <= 0.0 then
        begin
          SetErrCode(MatSing);
          Exit;
        end;

  { Dimension arrays }
  var U  : array[L..N, L..M] of Float;  { Matrix of indep. var. }
  var Z  : array[L..N] of Float;        { Dependent variable }
  var S1 : array[L..M] of Float;        { Singular values }
  var V1 : array[L..M, L..M] of Float;  { Matrix from SVD }

  if Mode = OLS then
    for I := L to N do
      begin
        K := I + Lb1 - L;
        Z[I] := Y[K];
        if ConsTerm then
          U[I,0] := 1.0;
        for J := 1 to M do
          U[I,J] := X[K,J];
      end
    else
      for I := L to N do
        begin
          K := I + Lb1 - L;
          Sigma := 1.0 / S[K];
          Z[I] := Y[K] * Sigma;
          if ConsTerm then
            U[I,0] := Sigma;
          for J := 1 to M do
            U[I,J] := X[K,J] * Sigma;
        end;

  { Perform singular value decomposition }
  SV_Decomp(U, S1, V1);

  if MathErr = MatOk then
    begin
      { Set the lowest singular values to zero }
      SV_SetZero(S1, SVDTol);

      { Solve the system }
      SV_Solve(U, S1, B, V1, Z);

      { Compute variance-covariance matrix }
      for I := L to M do
        for J := L to I do
          begin
            V[I, J] := 0.0;
            for K := L to M do
              if S1[K] > 0.0 then
                V[I, J] := V[I, J] + V1[I,K] * V1[J,K] / Sqr(S1[K]);
            V[J ,I] := V[I, J];
          end;
    end;
end;

procedure SVDFit(var X    : Matrix;
                 var Y    : Vector;
                 ConsTerm : Boolean;
                 SVDTol   : Float;
                 var B    : BVector;
                 var V    : VMatrix);
var
  S : array[Lb..Ub] of Float;  { Dummy variable }
begin
  GenSVDFit(OLS, X, Y, S, ConsTerm, SVDTol, B, V);
end;

procedure WSVDFit(var X    : Matrix;
                  var Y, S : Vector;
                  ConsTerm : Boolean;
                  SVDTol   : Float;
                  var B    : BVector;
                  var V    : VMatrix);
begin
  GenSVDFit(WLS, X, Y, S, ConsTerm, SVDTol, B, V);
end;

end.
